@@ -7,7 +7,7 @@ from dateutil import relativedelta
 from dateutil import rrule
 
 from trytond.model import Workflow, ModelSQL, ModelView, fields
-from trytond.pyson import Eval, Bool
+from trytond.pyson import Eval, Bool, If
 from trytond.pool import Pool
 from trytond.transaction import Transaction
 from trytond.wizard import Wizard, StateView, StateTransition, Button
@@ -38,13 +38,15 @@ class Asset(Workflow, ModelSQL, ModelView):
     supplier_invoice_line = fields.Many2One('account.invoice.line',
         'Supplier Invoice Line',
         domain=[
-            ('product', '=', Eval('product', -1)),
+            If(Eval('product', None) == None,
+                ('product', '=', -1),
+                ('product', '=', Eval('product', -1)),
+                ),
             ('invoice.type', '=', 'in_invoice'),
             ],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
-        on_change=['supplier_invoice_line', 'unit'],
         depends=['product', 'state'])
     customer_invoice_line = fields.Function(fields.Many2One(
             'account.invoice.line', 'Customer Invoice Line'),
@@ -62,8 +64,8 @@ class Asset(Workflow, ModelSQL, ModelView):
             },
         depends=['state'],
         required=True)
-    currency_digits = fields.Function(fields.Integer('Currency Digits',
-            on_change_with=['company']), 'on_change_with_currency_digits')
+    currency_digits = fields.Function(fields.Integer('Currency Digits'),
+        'on_change_with_currency_digits')
     quantity = fields.Float('Quantity',
         digits=(16, Eval('unit_digits', 2)),
         states={
@@ -73,14 +75,13 @@ class Asset(Workflow, ModelSQL, ModelView):
             },
         depends=['state', 'unit_digits'])
     unit = fields.Many2One('product.uom', 'Unit',
-        on_change_with=['product'],
         states={
             'readonly': (Bool(Eval('product'))
                 | (Eval('state') != 'draft')),
             },
         depends=['state'])
-    unit_digits = fields.Function(fields.Integer('Unit Digits',
-            on_change_with=['unit']), 'on_change_with_unit_digits')
+    unit_digits = fields.Function(fields.Integer('Unit Digits'),
+        'on_change_with_unit_digits')
     value = fields.Numeric('Value',
         digits=(16, Eval('currency_digits', 2)),
         states={
@@ -108,7 +109,6 @@ class Asset(Workflow, ModelSQL, ModelView):
         required=True,
         depends=['state'])
     end_date = fields.Date('End Date',
-        on_change_with=['end_date', 'product', 'start_date'],
         states={
             'readonly': (Eval('lines', [0]) | (Eval('state') != 'draft')),
             },
@@ -210,11 +210,13 @@ class Asset(Workflow, ModelSQL, ModelView):
             return journals[0].id
         return None
 
+    @fields.depends('company')
     def on_change_with_currency_digits(self, name=None):
         if self.company:
             return self.company.currency.digits
         return 2
 
+    @fields.depends('supplier_invoice_line', 'unit')
     def on_change_supplier_invoice_line(self):
         pool = Pool()
         Currency = pool.get('currency.currency')
@@ -236,13 +238,14 @@ class Asset(Workflow, ModelSQL, ModelView):
                     invoice.currency)
         else:
             new_values['value'] = invoice_line.amount
-        new_values['purchase_date'] = invoice.invoice_date
-        new_values['start_date'] = invoice.invoice_date
-        if invoice_line.product.depreciation_duration:
-            duration = relativedelta.relativedelta(
-                months=int(invoice_line.product.depreciation_duration),
-                days=-1)
-            new_values['end_date'] = new_values['start_date'] + duration
+        if invoice.invoice_date:
+            new_values['purchase_date'] = invoice.invoice_date
+            new_values['start_date'] = invoice.invoice_date
+            if invoice_line.product.depreciation_duration:
+                duration = relativedelta.relativedelta(
+                    months=int(invoice_line.product.depreciation_duration),
+                    days=-1)
+                new_values['end_date'] = new_values['start_date'] + duration
 
         if not self.unit:
             new_values['quantity'] = invoice_line.quantity
@@ -253,16 +256,19 @@ class Asset(Workflow, ModelSQL, ModelView):
 
         return new_values
 
+    @fields.depends('product')
     def on_change_with_unit(self):
         if not self.product:
             return None
         return self.product.default_uom.id
 
+    @fields.depends('unit')
     def on_change_with_unit_digits(self, name=None):
         if not self.unit:
             return 2
         return self.unit.digits
 
+    @fields.depends('end_date', 'product', 'start_date')
     def on_change_with_end_date(self):
         if all(getattr(self, k, None) for k in ('product', 'start_date')):
             if self.product.depreciation_duration:
@@ -296,7 +302,7 @@ class Asset(Workflow, ModelSQL, ModelView):
         # dateutil >= 2.0 has replace __nonzero__ by __bool__ which doesn't
         # work in Python < 3
         if delta == relativedelta.relativedelta():
-            return []
+            return [self.end_date]
         if self.frequency == 'monthly':
             rule = rrule.rrule(rrule.MONTHLY, dtstart=self.start_date,
                 bymonthday=-1)
@@ -320,7 +326,10 @@ class Asset(Workflow, ModelSQL, ModelView):
                     - relativedelta.relativedelta(days=1)]
                 + [l.date for l in self.lines])
             first_delta = dates[0] - start_date
-            last_delta = dates[-1] - dates[-2]
+            if len(dates) > 1:
+                last_delta = dates[-1] - dates[-2]
+            else:
+                last_delta = first_delta
             if self.frequency == 'monthly':
                 _, first_ndays = calendar.monthrange(
                     dates[0].year, dates[0].month)
